#!/usr/bin/env python
""" Tools for generating and interpolating between waveform cycles. """

import numpy as np

from osc_gen import dsp


class SigGen(object):
    """ Signal Generator """

    def __init__(self, harmonic=0):
        """ Init """

        self.num_points = 128
        self.amp = 1.0
        self.offset = 0.0
        self.harmonic = harmonic
        self.phase = 0

    def __base(self):
        """ Generate the base waveform cycle, a sawtooth or ramp from -1 to 1
        """

        repeats = self.harmonic + 1
        normalized_phase = self.phase / (2 * np.pi)
        start = normalized_phase
        stop = start + repeats

        wave = np.linspace(start, stop, num=self.num_points, dtype=np.float32)

        # wrap and shift to +/- 1
        wrap_threshold = np.finfo(np.float32).eps
        wave %= 1 + wrap_threshold
        wave *= 2
        wave -= 1

        return wave

    def saw(self):
        """ Generate a sawtooth wave cycle """

        return self.amp * self.__base()

    def tri(self):
        """ Generate a triangle wave cycle """

        return self.amp * (np.abs(self.__base()) * -2 + 1)

    def pls(self, width):
        """ Generate a pulse wave cycle

            @param width float : Pulse width or duty cycle, between -1 and 1,
                where 0 corresponds to a square wave.
        """

        p = self.__base()
        p[np.where(p < width)[0]] = -1.
        p[np.where(p >= width)[0]] = 1.

        return self.amp * p

    def sqr(self):
        """ Generate a square wave cycle """

        return self.pls(0)

    def sin(self):
        """ Generate a sine wave cycle """

        return self.amp * np.sin(np.pi * (self.__base() + 0.5))

    def arb(self, data):
        """ Generate an arbitrary wave cycle. The provided data will be
        interpolated, if possible, to occupy the correct number of samples for
        a single cycle at our reference frequency and then normalized and
        scaled as appropriate.

        @param data seq : A sequence of samples representing a single cycle
            of a wave
        """

        interp_y = data
        num = interp_y.size
        interp_x = np.linspace(0, num, num=num)
        interp_xx = np.linspace(0, num, num=self.num_points)
        interp_yy = np.interp(interp_xx, interp_x, interp_y)
        dsp.normalize(interp_yy)

        return interp_yy


def morph(waves, new_num):
    """ Take a number of wave cycles and generate a higher number of wave cycles
        where the original waves are linearly interpolated from one to the next
        to fill in the gaps.

        @param waves sequence : A sequence of wave cycles
        @param new_num int : The reuqired number of wave cycles in the new
            seuqence
    """

    inp = list(waves)
    inp_num = len(inp)

    if inp_num >= new_num:
        msg = "Can't morph a group into a smaller or equal group ({0} to {1})"
        raise ValueError(msg.format(inp_num, new_num))

    if inp_num < 2:
        msg = "Can't morph between less than 2 signals ({0})"
        raise ValueError(msg.format(inp_num))

    if inp_num == 2:
        return __morph_two(inp[0], inp[1], new_num)

    ranges = __detrmine_morph_ranges(inp_num, new_num)

    return __morph_many(inp, ranges)


def __detrmine_morph_ranges(inp_num, new_num):
    """ Find a set of integer gaps sizes between two set sizes

        @param inp_num int : The original set size
        @param new_num int : The new set size
    """

    gap_num = inp_num - 1
    gap_val = int(round(float(new_num) / gap_num, 0))
    ranges = [gap_val if j is 0 else gap_val + 1 for j in range(gap_num)]
    k = -1
    while sum(ranges) < new_num + gap_num - 1:
        ranges[k] += 1
        k -= 1
    while sum(ranges) > new_num + gap_num - 1:
        ranges[k] -= 1
        k -= 1

    return ranges


def __morph_many(waves, gaps):
    """ Morph between more then two sequences

        @param waves sequence : A sequence of wave cycles
        @param gaps sequence : The size of the gap between each pair of cycles
    """

    morphed = []
    prev_wave = None
    i = 0
    for curr_wave in waves:
        if prev_wave is not None:
            if i is 0:
                start = 0
            else:
                start = 1

            morphed.extend(
                [x for x in __morph_two(
                    prev_wave, curr_wave, gaps[i])][start:])
            i += 1

        prev_wave = curr_wave

    return morphed


def __morph_two(wave_one, wave_two, new_num):
    """ Morph between two wave cycles.

        @param a sequence : The first wave cycle
        @param b sequence : The second wave cycle
        @param n int : The reuqired number of wave cycles in the new seuqence
    """

    alphas = (s / (new_num - 1.0) for s in range(new_num))

    return [wave_one * (1 - m) + wave_two * m for m in alphas]



#!/usr/bin/env python
""" Zebra oscillator waves """

from __future__ import division
from __future__ import print_function
import numpy as np

from osc_gen import wavfile
from osc_gen import dsp
from osc_gen import sig


class WaveTable(object):
    """ An n-slot wavetable """

    def __init__(self, waves=None, num_waves=16, wave_len=128):
        """
        Init

        @param waves sequence : A sequence of numpy arrays containing wave
            data to form the wavetable
        """

        self.num_waves = num_waves
        self.wave_len = wave_len

        if waves is None:
            self.waves = []
        else:
            self.waves = waves

    def clear(self):
        """ Clear the wavetable so that all slots contain zero """

        self.waves = []

    def get_wave_at_index(self, index):
        """
        Get the wave at a specific slot index

        @param index int : The slot index to get the wave from

        @returns np.ndarray : Wavefore at given index
        """

        if index >= len(self.waves):
            return np.zeros(self.wave_len)
        else:
            return self.waves[index]

    def get_waves(self):
        """ Get all of the waves in the table """

        for i in range(self.num_waves):
            yield self.get_wave_at_index(i)

    def from_wav(self, filename, sig_gen=None, resynthesize=False):
        """
        Populate the wavetable from a wav file by filling all slots with
        cycles from a wav file.

        @param filename str : Wav file name.
        @param sig_gen SigGen : SigGen to use for regenerating the signal.
        @param resynthesize : If True, the signal is resynthesised using the
            harmonic series of the original signal - works best on signals with
            a low findamental frequency (< 200 Hz). If False, n evenly spaced
            single cycles are extracted from the input (default False).

        @returns WaveTable : self, populated by content from the wav file
        settings as this one
        """

        a, fs = wavfile.read(filename, with_sample_rate=True)

        if sig_gen is None:
            sig_gen = sig.SigGen()

        if resynthesize:

            num_sections = self.num_waves

            while 1:
                a = a[:len(a) - (len(a) % num_sections)]
                sections = np.split(a, num_sections)
                try:
                    self.waves = [dsp.resynthesize(s, sig_gen) for s in sections]
                    break
                except dsp.NotEnoughSamplesError as e:
                    num_sections -= 1
                    if num_sections <= 0:
                        raise e

            if num_sections < self.num_waves:
                self.waves = sig.morph(self.waves, self.num_waves)

        else:
            cycles = dsp.slice_cycles(a, self.num_waves, fs)
            self.waves = [sig_gen.arb(c) for c in cycles]

        return self



#!/usr/bin/env python
""" Plot waveforms """

import matplotlib.pyplot as plt


def plot_wave(wave):
    """ Plot a single wave """

    plt.plot(wave)
    plt.show()
    plt.gcf().clear()


def plot_wavetable(wavetable, title=''):
    """ Plot all waves in a wavetable """

    for wave in wavetable.waves:
        plt.plot(wave)

    plt.title(title)
    plt.show()
    plt.gcf().clear()



#!/usr/bin/env python
""" Zebra2 Oscillator wavetable """


def write_wavetable(wavetable, filename):
    """ Write wavetable to an h2p oscillator file

        @param wavetable zwave.WaveTable : Wavetable
        @param filename str : File name to write to
    """

    table_size = wavetable.wave_len

    if table_size is None:
        return

    with open(filename, 'w') as osc_file:

        osc_file.write("#defaults=no\n")
        osc_file.write("#cm=OSC\n")
        osc_file.write("Wave=2\n")
        osc_file.write("<?\n")
        osc_file.write("\n")

        osc_file.write("float Wave[")
        osc_file.write(str(table_size))
        osc_file.write("];\n")
        osc_file.write("\n")

        for i, wave in enumerate(wavetable.get_waves()):

            if wave is None:
                continue

            wave_num = i + 1
            # scale to avoid overflow resulting from finite precision
            scaled_wave = wave * 0.999969
            osc_file.write("//table ")
            osc_file.write(str(wave_num))
            osc_file.write("\n")

            for index, value in enumerate(scaled_wave):
                osc_file.write("Wave[")
                osc_file.write(str(index))
                osc_file.write("] = ")
                osc_file.write('{0:.10f}'.format(value))
                osc_file.write(";\n")

            osc_file.write("Selected.WaveTable.set(")
            osc_file.write(str(wave_num))
            osc_file.write(", Wave);\n")
            osc_file.write("\n")

        osc_file.write("?>")



#!/usr/bin/env python
""" Extend wavfile to read data as normalized float """

try:
    import soundfile as sf
    HAS_SOUNDFILE = True
except ImportError:
    HAS_SOUNDFILE = False

import wave
import struct
import numpy as np


def _float_to_ibytes(vals):
    """ Convert a sequence of vals to 16-bit bytes """

    num = len(vals)
    afloats = np.array(vals)
    afloats = (afloats * 32768).astype('int')
    np.clip(afloats, -32768, 32767, out=afloats)
    return struct.pack('<{0}h'.format(num), *(aval for aval in afloats))


def _ibytes_to_float(vals):

    num = len(vals)
    num = int(num / 4)
    afloats = struct.unpack('<{0}h'.format(num), vals)
    afloats = np.array(afloats).astype(float)
    afloats /= 32768.
    return afloats.astype(float)


def _read_using_wave(filename):

    wave_file = wave.open(filename, 'r')

    if wave_file.getnchannels() != 1:
        raise ValueError("only mono supported")

    if wave_file.getsampwidth() != 2:
        raise ValueError("only 16 bit supported")

    num_frames = wave_file.getnframes()
    data = _ibytes_to_float(wave_file.readframes(num_frames))
    fs = wave_file.getframerate()

    return data, fs


def read(filename, with_sample_rate=False):
    """ Read wav file and convert to normalized float """

    if HAS_SOUNDFILE:
        data, fs = sf.read(filename)
        # take only the first channel of the audio
        if len(data.shape) > 1:
            data = np.swapaxes(data, 0, 1)[0]
    else:
        data, fs = _read_using_wave(filename)

    # center on 0
    data = data.astype(float)
    data -= np.mean(data)

    # normalize to +/- 1.0
    data_max = max(abs(data))
    data /= data_max

    if with_sample_rate:
        return data, fs
    else:
        return data


def write(data, filename, samplerate=44100):
    """ Write wav file """

    wave_file = wave.open(filename, 'w')
    wave_file.setframerate(samplerate)
    wave_file.setnchannels(1)
    wave_file.setsampwidth(2)
    wave_file.writeframes(_float_to_ibytes(data))
    wave_file.close()


def write_wavetable(wavetable, filename, samplerate=44100):
    """ Write wavetable to file """

    wave_file = wave.open(filename, 'w')
    wave_file.setframerate(samplerate)
    wave_file.setnchannels(1)
    wave_file.setsampwidth(2)
    for wt_wave in wavetable.get_waves():
        wave_file.writeframes(_float_to_ibytes(wt_wave))
    wave_file.close()



#!/usr/bin/env python
""" Setup osc_gen """

import os
import sys

from setuptools import setup
from setuptools.command.install import install

VERSION = "0.1.1"

LONG_DESCRIPTION = (
    """osc_gen is a Python library for creating and managing oscillator wavetables.

Functionality includes:

* Generating common waveforms (sine, saw, square, etc.)
* Oscillator effects (waveshaping, distortion, downsampling, etc.)
* Resynthesising or slicing audio from wav files or other sources
* Saving wavetables to a wav file for use in samplers
* Saving wavetables in .h2p format for use in the u-he Zebra2 synthesiser"""
)


class VerifyVersionCommand(install):
    """ Custom command to verify that the git tag matches our version """
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


setup(
    name='osc_gen',
    version=VERSION,
    description='Generate oscilator wavetoables',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    author='Harvey Ormston',
    author_email='harveyormston@me.com',
    url='https://github.com/harveyormston/osc_gen',
    license='GPL3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Topic :: Multimedia :: Sound/Audio :: Sound Synthesis',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'],
    keywords='audio dsp synthsis',
    project_urls={
        'Source': 'https://github.com/harveyormston/osc_gen',
    },
    packages=['osc_gen'],
    install_requires=[
        "numpy>=1.11.3",
        "scipy>=0.18.1",
        "pysoundfile"],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4',
    cmdclass={'verify': VerifyVersionCommand}
)



#!/usr/bin/env python
""" Process waveforms """

import math
from copy import deepcopy

import numpy as np


class NotEnoughSamplesError(Exception):
    """ Not Enough Samples """
    pass


def normalize(inp):
    """ Normalize a signal to the range +/- 1

        @param inp seq : A sequence of samples
    """

    inp -= (np.amax(inp) + np.amin(inp)) / 2
    amp = np.amax(np.absolute(inp))

    if amp > 0:
        inp /= np.amax(abs(inp))

    return inp


def clip(inp, amount, bias=0):
    """ Hard-clip a signal

        @param inp seq : A sequence of samples
        @param amount number : Amount of clipping
        @param bias number : Pre-distortion DC bias
    """

    gain = 1 + amount

    inp += bias
    inp *= gain
    np.clip(inp, -1., 1., out=inp)

    return normalize(inp)


def tube(inp, amount, bias=0):
    """ Tube saturate a signal

        @param inp seq : A sequence of samples
        @param amount number : Amount of distortion
        @param bias number : Pre-distortion DC bias
    """

    gain = 1 + amount
    inp += bias
    inp *= gain
    for i, val in enumerate(inp):
        inp[i] = math.exp(-np.logaddexp(0, -val))

    inp *= 2
    inp -= 1

    return normalize(inp)


def fold(inp, amount, bias=0):
    """ Perform wave folding

        @param inp seq : A sequence of samples
        @param amount number : Amount of distortion
        @param bias number : Pre-distortion DC bias
    """

    gain = 1 + amount
    inp += bias
    inp *= gain
    while (max(abs(inp))) > 1:
        for i, val in enumerate(inp):
            if val > 1:
                inp[i] = 2 - val
            if val < -1:
                inp[i] = -2 - val

    return normalize(inp)


def shape(inp, amount=1, bias=0, power=3):
    """ Perform polynomial waveshaping

        @param inp seq : A sequence of samples
        @param amount number : Amount of shaping
            (1: maximum shaping, 0: no shaping)
        @param bias number : Pre-distortion DC bias
        @param power number : Polynomial power
    """

    biased = inp + bias

    # make another copy to apply polynomial shaping to the biased input
    shaped = np.empty_like(biased)
    # shape positive and negative halves of the signal symmetrically
    shaped[biased >= 0] = np.power(biased[biased >= 0], power) * amount
    shaped[biased < 0] = -np.power(-biased[biased < 0], power) * amount
    # de-bais
    shaped -= bias
    normalize(shaped)

    inp *= (1 - amount)
    inp += shaped * amount

    return normalize(inp)


def slew(inp, rate, inv=False):
    """ Apply slew or overhoot to a signal. Slew smooths steep transients in
        the signal while overshoot results in a sharper transient with
        ringing.

        @param rate float : Slew rate, between 0 and 1
        @param inp seq : A sequence of samples
        @param inv bool : If True, overshoot will be applied. if False,
                          slew will be applied. (default=False).
    """

    prev = 0.
    # process twice in order to allow for settling
    for num in range(2):
        for i, val in enumerate(inp):
            if inv:
                curr = prev * (rate - 1.) + (val * rate)
            else:
                curr = val * (1. - rate) + (prev * rate)
            prev = curr
            if num > 0:
                inp[i] = curr

    return normalize(inp)


def downsample(inp, factor):
    """ Reduce the effective sample rate of a signal, resulting in aliasing.

        @param inp seq : A sequence of samples
        @param factor int : Downsampling factor
    """

    if factor < 1:
        raise ValueError(
            "Downsampling factor ({0}) cannot be < 1".format(factor))

    if factor == 1:
        return inp
    else:
        # the aliasing is deliberate!
        for i, val in enumerate(inp):
            if i % factor == 0:
                last = val
            else:
                inp[i] = last

    return normalize(inp)


def quantize(inp, depth):
    """ Reduce the bit depth of a signal.

        @param inp seq : A sequence of samples
        @param depth number : New bit depth in bits
    """

    scale = 2 ** depth - 1

    for i, val in enumerate(inp):
        if val > 0:
            inp[i] = np.ceil(val * scale) / scale
        elif val < 0:
            inp[i] = np.floor(val * scale) / scale

    return normalize(inp)


def fundamental(inp, fs):
    """ Find the fundamental frequency in Hz of a given input """

    h = np.hamming(len(inp))
    w = np.fft.fft(inp * h)
    f = np.fft.fftfreq(len(w))
    i = np.argmax(np.abs(w))

    return abs(f[i] * fs)


def harmonic_series(inp):
    """ Find the harmonic series of a periodic input """

    L = min(64, len(inp) // 501)
    M = 501 * L
    if len(inp) < M:
        raise NotEnoughSamplesError("Got {0} samples, need at least {1}.".format(len(inp), M))

    # produce symmetrical, windowed fft
    hM1 = int(np.floor((M + 1) / 2))
    hM2 = int(np.floor(M / 2))
    x1 = inp[:M] * np.hamming(M)
    N = 1024 * L
    buf = np.zeros(N)
    buf[:hM1] = x1[hM2:]
    buf[N - hM2:] = x1[:hM2]
    fft = np.fft.fft(buf)[:N // 2]

    # peak amplitude assumed to be fundamental frequency
    i_fund = np.argmax(np.abs(fft))

    # get fft components from only the harmonics, harmonics are picked by
    # taking the value with the highest amplitude around each harmonic
    # frequency
    ws = i_fund // 4
    hs = np.array(
        [fft[i - ws:i + ws][np.abs(fft[i - ws:i + ws]).argmax()]
         for i in range(i_fund, N // 2, i_fund)])

    # normalize magnitude and phase
    hs_amp = np.abs(hs)
    hs_ang = np.angle(hs)
    hs = hs_amp * np.exp(1j * (hs_ang - hs_ang[0])) / hs_amp[0]

    return hs


def slice_cycles(inp, n, fs):
    """ Extact n single-cycle slices from a signal """

    def nearest(arr, val):
        """ find the nearest value in an array to a given value """
        return arr[np.argmin(np.abs(arr - val))]

    zero_crossings = np.where(np.diff(np.sign(inp)) > 0)[0] + 1

    if len(zero_crossings) < 1:
        raise ValueError("No zero crossings found.")

    freq = fundamental(inp, fs)
    samples_per_cycle = fs / freq
    end = len(inp) - samples_per_cycle

    slots = np.linspace(0, end, n)
    slots = np.around(slots).astype(int)
    slots = np.unique([nearest(zero_crossings, slot) for slot in slots])

    return [inp[x:x + int(samples_per_cycle)] for x in slots]


def resynthesize(inp, sig_gen):
    """
    Resynthesize a signal from its harmonic series

    @param sig_gen SigGen : SigGen to use for regenerating the signal.
    """

    sg = deepcopy(sig_gen)
    max_harmonic = sig_gen.num_points // 2
    hs = harmonic_series(inp)
    s = np.zeros(sg.num_points)

    for i, h in enumerate(hs):

        sg.harmonic = i
        sg.amp = np.abs(h)
        sg.phase = np.angle(h)

        s += sg.sin()

        if i >= max_harmonic:
            break

    normalize(s)

    return s
